@@ -11,6 +11,7 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+//go:build go1.14
 // +build go1.14
 
 package web
@@ -20,15 +21,24 @@ import (
 	"crypto/x509"
 	"errors"
 	"fmt"
-	"io/ioutil"
+	"io"
 	"net"
 	"net/http"
+	"os"
 	"regexp"
 	"sync"
 	"testing"
 	"time"
 )
 
+// Helpers for literal FlagConfig
+func OfBool(i bool) *bool {
+	return &i
+}
+func OfString(i string) *string {
+	return &i
+}
+
 var (
 	port       = getPort()
 	testlogger = &testLogger{}
@@ -52,7 +62,11 @@ var (
 		"Unknown curve":                regexp.MustCompile(`unknown curve`),
 		"Unknown TLS version":          regexp.MustCompile(`unknown TLS version`),
 		"No HTTP2 cipher":              regexp.MustCompile(`TLSConfig.CipherSuites is missing an HTTP/2-required`),
-		"Incompatible TLS version":     regexp.MustCompile(`protocol version not supported`),
+		// The first token is returned by Go <= 1.17 and the second token is returned by Go >= 1.18.
+		"Incompatible TLS version": regexp.MustCompile(`protocol version not supported|no supported versions satisfy MinVersion and MaxVersion`),
+		"Bad certificate":          regexp.MustCompile(`bad certificate`),
+		"Invalid value":            regexp.MustCompile(`invalid value for`),
+		"Invalid header":           regexp.MustCompile(`HTTP header ".*" can not be configured`),
 	}
 )
 
@@ -84,6 +98,7 @@ type TestInputs struct {
 	CurvePreferences    []tls.CurveID
 	Username            string
 	Password            string
+	ClientCertificate   string
 }
 
 func TestYAMLFiles(t *testing.T) {
@@ -95,77 +110,77 @@ func TestYAMLFiles(t *testing.T) {
 		},
 		{
 			Name:           `empty config yml`,
-			YAMLConfigPath: "testdata/tls_config_empty.yml",
+			YAMLConfigPath: "testdata/web_config_empty.yml",
 			ExpectedError:  nil,
 		},
 		{
 			Name:           `invalid config yml (invalid structure)`,
-			YAMLConfigPath: "testdata/tls_config_junk.yml",
+			YAMLConfigPath: "testdata/web_config_junk.yml",
 			ExpectedError:  ErrorMap["YAML error"],
 		},
 		{
 			Name:           `invalid config yml (invalid key)`,
-			YAMLConfigPath: "testdata/tls_config_junk_key.yml",
+			YAMLConfigPath: "testdata/web_config_junk_key.yml",
 			ExpectedError:  ErrorMap["YAML error"],
 		},
 		{
 			Name:           `invalid config yml (cert path empty)`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_certPath_empty.bad.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_certPath_empty.bad.yml",
 			ExpectedError:  ErrorMap["Invalid CertPath"],
 		},
 		{
 			Name:           `invalid config yml (key path empty)`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_keyPath_empty.bad.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_keyPath_empty.bad.yml",
 			ExpectedError:  ErrorMap["Invalid KeyPath"],
 		},
 		{
 			Name:           `invalid config yml (cert path and key path empty)`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_certPath_keyPath_empty.bad.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_certPath_keyPath_empty.bad.yml",
 			ExpectedError:  ErrorMap["Invalid CertPath"],
 		},
 		{
 			Name:           `invalid config yml (cert path invalid)`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_certPath_invalid.bad.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_certPath_invalid.bad.yml",
 			ExpectedError:  ErrorMap["No such file"],
 		},
 		{
 			Name:           `invalid config yml (key path invalid)`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_keyPath_invalid.bad.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_keyPath_invalid.bad.yml",
 			ExpectedError:  ErrorMap["No such file"],
 		},
 		{
 			Name:           `invalid config yml (cert path and key path invalid)`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_certPath_keyPath_invalid.bad.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_certPath_keyPath_invalid.bad.yml",
 			ExpectedError:  ErrorMap["No such file"],
 		},
 		{
 			Name:           `invalid config yml (invalid ClientAuth)`,
-			YAMLConfigPath: "testdata/tls_config_noAuth.bad.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth.bad.yml",
 			ExpectedError:  ErrorMap["ClientCA set without policy"],
 		},
 		{
 			Name:           `invalid config yml (invalid ClientCAs filepath)`,
-			YAMLConfigPath: "testdata/tls_config_auth_clientCAs_invalid.bad.yml",
+			YAMLConfigPath: "testdata/web_config_auth_clientCAs_invalid.bad.yml",
 			ExpectedError:  ErrorMap["No such file"],
 		},
 		{
 			Name:           `invalid config yml (invalid user list)`,
-			YAMLConfigPath: "testdata/tls_config_auth_user_list_invalid.bad.yml",
+			YAMLConfigPath: "testdata/web_config_auth_user_list_invalid.bad.yml",
 			ExpectedError:  ErrorMap["Bad password"],
 		},
 		{
 			Name:           `invalid config yml (bad cipher)`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_inventedCiphers.bad.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_inventedCiphers.bad.yml",
 			ExpectedError:  ErrorMap["Unknown cipher"],
 		},
 		{
 			Name:           `invalid config yml (bad curves)`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_inventedCurves.bad.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_inventedCurves.bad.yml",
 			ExpectedError:  ErrorMap["Unknown curve"],
 		},
 		{
 			Name:           `invalid config yml (bad TLS version)`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_wrongTLSVersion.bad.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_wrongTLSVersion.bad.yml",
 			ExpectedError:  ErrorMap["Unknown TLS version"],
 		},
 	}
@@ -190,56 +205,56 @@ func TestServerBehaviour(t *testing.T) {
 		},
 		{
 			Name:           `valid tls config yml and default client`,
-			YAMLConfigPath: "testdata/tls_config_noAuth.good.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth.good.yml",
 			ExpectedError:  ErrorMap["HTTP Request to HTTPS server"],
 		},
 		{
 			Name:           `valid tls config yml and tls client`,
-			YAMLConfigPath: "testdata/tls_config_noAuth.good.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth.good.yml",
 			UseTLSClient:   true,
 			ExpectedError:  nil,
 		},
 		{
 			Name:                `valid tls config yml with TLS 1.1 client`,
-			YAMLConfigPath:      "testdata/tls_config_noAuth.good.yml",
+			YAMLConfigPath:      "testdata/web_config_noAuth.good.yml",
 			UseTLSClient:        true,
 			ClientMaxTLSVersion: tls.VersionTLS11,
 			ExpectedError:       ErrorMap["Incompatible TLS version"],
 		},
 		{
 			Name:           `valid tls config yml with all ciphers`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_allCiphers.good.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_allCiphers.good.yml",
 			UseTLSClient:   true,
 			ExpectedError:  nil,
 		},
 		{
 			Name:           `valid tls config yml with some ciphers`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_someCiphers.good.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_someCiphers.good.yml",
 			UseTLSClient:   true,
 			CipherSuites:   []uint16{tls.TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256},
 			ExpectedError:  nil,
 		},
 		{
 			Name:           `valid tls config yml with no common cipher`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_someCiphers.good.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_someCiphers.good.yml",
 			UseTLSClient:   true,
 			CipherSuites:   []uint16{tls.TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA},
 			ExpectedError:  ErrorMap["Handshake failure"],
 		},
 		{
 			Name:           `valid tls config yml with multiple client ciphers`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_someCiphers.good.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_someCiphers.good.yml",
 			UseTLSClient:   true,
 			CipherSuites: []uint16{
+				tls.TLS_RSA_WITH_AES_128_CBC_SHA,
 				tls.TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256,
-				tls.TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384,
 			},
-			ActualCipher:  tls.TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384,
+			ActualCipher:  tls.TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256,
 			ExpectedError: nil,
 		},
 		{
 			Name:           `valid tls config yml with multiple client ciphers, client chooses cipher`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_someCiphers_noOrder.good.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_someCiphers_noOrder.good.yml",
 			UseTLSClient:   true,
 			CipherSuites: []uint16{
 				tls.TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256,
@@ -250,36 +265,88 @@ func TestServerBehaviour(t *testing.T) {
 		},
 		{
 			Name:           `valid tls config yml with all curves`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_allCurves.good.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_allCurves.good.yml",
 			UseTLSClient:   true,
 			ExpectedError:  nil,
 		},
 		{
 			Name:             `valid tls config yml with some curves`,
-			YAMLConfigPath:   "testdata/tls_config_noAuth_someCurves.good.yml",
+			YAMLConfigPath:   "testdata/web_config_noAuth_someCurves.good.yml",
 			UseTLSClient:     true,
 			CurvePreferences: []tls.CurveID{tls.CurveP521},
 			ExpectedError:    nil,
 		},
 		{
 			Name:             `valid tls config yml with no common curves`,
-			YAMLConfigPath:   "testdata/tls_config_noAuth_someCurves.good.yml",
+			YAMLConfigPath:   "testdata/web_config_noAuth_someCurves.good.yml",
 			UseTLSClient:     true,
 			CurvePreferences: []tls.CurveID{tls.CurveP384},
 			ExpectedError:    ErrorMap["Handshake failure"],
 		},
 		{
 			Name:           `valid tls config yml with non-http2 ciphers`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_noHTTP2.good.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_noHTTP2.good.yml",
 			UseTLSClient:   true,
 			ExpectedError:  nil,
 		},
 		{
 			Name:           `valid tls config yml with non-http2 ciphers but http2 enabled`,
-			YAMLConfigPath: "testdata/tls_config_noAuth_noHTTP2Cipher.bad.yml",
+			YAMLConfigPath: "testdata/web_config_noAuth_noHTTP2Cipher.bad.yml",
 			UseTLSClient:   true,
 			ExpectedError:  ErrorMap["No HTTP2 cipher"],
 		},
+		{
+			Name:           `valid tls config yml and tls client with RequireAnyClientCert`,
+			YAMLConfigPath: "testdata/tls_config_noAuth.requireanyclientcert.good.yml",
+			UseTLSClient:   true,
+			ExpectedError:  ErrorMap["Bad certificate"],
+		},
+		{
+			Name:           `valid headers config`,
+			YAMLConfigPath: "testdata/web_config_headers.good.yml",
+		},
+		{
+			Name:           `invalid X-Content-Type-Options headers config`,
+			YAMLConfigPath: "testdata/web_config_headers_content_type_options.bad.yml",
+			ExpectedError:  ErrorMap["Invalid value"],
+		},
+		{
+			Name:           `invalid X-Frame-Options headers config`,
+			YAMLConfigPath: "testdata/web_config_headers_frame_options.bad.yml",
+			ExpectedError:  ErrorMap["Invalid value"],
+		},
+		{
+			Name:           `HTTP header that can not be overridden`,
+			YAMLConfigPath: "testdata/web_config_headers_extra_header.bad.yml",
+			ExpectedError:  ErrorMap["Invalid header"],
+		},
+		{
+			Name:              `valid tls config yml and tls client with RequireAnyClientCert (present certificate)`,
+			YAMLConfigPath:    "testdata/tls_config_noAuth.requireanyclientcert.good.yml",
+			UseTLSClient:      true,
+			ClientCertificate: "client_selfsigned",
+			ExpectedError:     nil,
+		},
+		{
+			Name:           `valid tls config yml and tls client with RequireAndVerifyClientCert`,
+			YAMLConfigPath: "testdata/tls_config_noAuth.requireandverifyclientcert.good.yml",
+			UseTLSClient:   true,
+			ExpectedError:  ErrorMap["Bad certificate"],
+		},
+		{
+			Name:              `valid tls config yml and tls client with RequireAndVerifyClientCert (present certificate)`,
+			YAMLConfigPath:    "testdata/tls_config_noAuth.requireandverifyclientcert.good.yml",
+			UseTLSClient:      true,
+			ClientCertificate: "client_selfsigned",
+			ExpectedError:     nil,
+		},
+		{
+			Name:              `valid tls config yml and tls client with RequireAndVerifyClientCert (present wrong certificate)`,
+			YAMLConfigPath:    "testdata/tls_config_noAuth.requireandverifyclientcert.good.yml",
+			UseTLSClient:      true,
+			ClientCertificate: "client2_selfsigned",
+			ExpectedError:     ErrorMap["Bad certificate"],
+		},
 	}
 	for _, testInputs := range testTables {
 		t.Run(testInputs.Name, testInputs.Test)
@@ -300,8 +367,8 @@ func TestConfigReloading(t *testing.T) {
 		}
 	}()
 
-	goodYAMLPath := "testdata/tls_config_noAuth.good.yml"
-	badYAMLPath := "testdata/tls_config_noAuth.good.blocking.yml"
+	goodYAMLPath := "testdata/web_config_noAuth.good.yml"
+	badYAMLPath := "testdata/web_config_noAuth.good.blocking.yml"
 
 	server := &http.Server{
 		Addr: port,
@@ -319,11 +386,16 @@ func TestConfigReloading(t *testing.T) {
 				recordConnectionError(errors.New("Panic starting server"))
 			}
 		}()
-		err := Listen(server, badYAMLPath, testlogger)
+		flagsBadYAMLPath := FlagConfig{
+			WebListenAddresses: &([]string{port}),
+			WebSystemdSocket:   OfBool(false),
+			WebConfigFile:      OfString(badYAMLPath),
+		}
+		err := Listen(server, &flagsBadYAMLPath, testlogger)
 		recordConnectionError(err)
 	}()
 
-	client := getTLSClient()
+	client := getTLSClient("")
 
 	TestClientConnection := func() error {
 		time.Sleep(250 * time.Millisecond)
@@ -331,7 +403,7 @@ func TestConfigReloading(t *testing.T) {
 		if err != nil {
 			return err
 		}
-		body, err := ioutil.ReadAll(r.Body)
+		body, err := io.ReadAll(r.Body)
 		if err != nil {
 			return err
 		}
@@ -377,7 +449,6 @@ func (test *TestInputs) Test(t *testing.T) {
 	}()
 
 	server := &http.Server{
-		Addr: port,
 		Handler: http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
 			w.Write([]byte("Hello World!"))
 		}),
@@ -389,7 +460,12 @@ func (test *TestInputs) Test(t *testing.T) {
 				recordConnectionError(errors.New("Panic starting server"))
 			}
 		}()
-		err := ListenAndServe(server, test.YAMLConfigPath, testlogger)
+		flags := FlagConfig{
+			WebListenAddresses: &([]string{port}),
+			WebSystemdSocket:   OfBool(false),
+			WebConfigFile:      &test.YAMLConfigPath,
+		}
+		err := ListenAndServe(server, &flags, testlogger)
 		recordConnectionError(err)
 	}()
 
@@ -397,7 +473,7 @@ func (test *TestInputs) Test(t *testing.T) {
 		var client *http.Client
 		var proto string
 		if test.UseTLSClient {
-			client = getTLSClient()
+			client = getTLSClient(test.ClientCertificate)
 			t := client.Transport.(*http.Transport)
 			t.TLSClientConfig.MaxVersion = test.ClientMaxTLSVersion
 			if len(test.CipherSuites) > 0 {
@@ -432,14 +508,14 @@ func (test *TestInputs) Test(t *testing.T) {
 			if r.TLS.CipherSuite != test.ActualCipher {
 				recordConnectionError(
 					fmt.Errorf("bad cipher suite selected. Expected: %s, got: %s",
-						tls.CipherSuiteName(r.TLS.CipherSuite),
 						tls.CipherSuiteName(test.ActualCipher),
+						tls.CipherSuiteName(r.TLS.CipherSuite),
 					),
 				)
 			}
 		}
 
-		body, err := ioutil.ReadAll(r.Body)
+		body, err := io.ReadAll(r.Body)
 		if err != nil {
 			recordConnectionError(err)
 			return
@@ -489,11 +565,23 @@ func (test *TestInputs) isCorrectError(returnedError error) bool {
 	return true
 }
 
-func getTLSClient() *http.Client {
-	cert, err := ioutil.ReadFile("testdata/tls-ca-chain.pem")
+func getTLSClient(clientCertName string) *http.Client {
+	cert, err := os.ReadFile("testdata/tls-ca-chain.pem")
 	if err != nil {
 		panic("Unable to start TLS client. Check cert path")
 	}
+
+	var clientCertficate tls.Certificate
+	if clientCertName != "" {
+		clientCertficate, err = tls.LoadX509KeyPair(
+			"testdata/"+clientCertName+".pem",
+			"testdata/"+clientCertName+".key",
+		)
+		if err != nil {
+			panic(fmt.Sprintf("failed to load client certificate: %v", err))
+		}
+	}
+
 	client := &http.Client{
 		Transport: &http.Transport{
 			TLSClientConfig: &tls.Config{
@@ -502,6 +590,9 @@ func getTLSClient() *http.Client {
 					caCertPool.AppendCertsFromPEM(cert)
 					return caCertPool
 				}(),
+				GetClientCertificate: func(req *tls.CertificateRequestInfo) (*tls.Certificate, error) {
+					return &clientCertficate, nil
+				},
 			},
 		},
 	}
@@ -509,19 +600,19 @@ func getTLSClient() *http.Client {
 }
 
 func swapFileContents(file1, file2 string) error {
-	content1, err := ioutil.ReadFile(file1)
+	content1, err := os.ReadFile(file1)
 	if err != nil {
 		return err
 	}
-	content2, err := ioutil.ReadFile(file2)
+	content2, err := os.ReadFile(file2)
 	if err != nil {
 		return err
 	}
-	err = ioutil.WriteFile(file1, content2, 0644)
+	err = os.WriteFile(file1, content2, 0644)
 	if err != nil {
 		return err
 	}
-	err = ioutil.WriteFile(file2, content1, 0644)
+	err = os.WriteFile(file2, content1, 0644)
 	if err != nil {
 		return err
 	}
@@ -532,25 +623,25 @@ func TestUsers(t *testing.T) {
 	testTables := []*TestInputs{
 		{
 			Name:           `without basic auth`,
-			YAMLConfigPath: "testdata/tls_config_users_noTLS.good.yml",
+			YAMLConfigPath: "testdata/web_config_users_noTLS.good.yml",
 			ExpectedError:  ErrorMap["Unauthorized"],
 		},
 		{
 			Name:           `with correct basic auth`,
-			YAMLConfigPath: "testdata/tls_config_users_noTLS.good.yml",
+			YAMLConfigPath: "testdata/web_config_users_noTLS.good.yml",
 			Username:       "dave",
 			Password:       "dave123",
 			ExpectedError:  nil,
 		},
 		{
 			Name:           `without basic auth and TLS`,
-			YAMLConfigPath: "testdata/tls_config_users.good.yml",
+			YAMLConfigPath: "testdata/web_config_users.good.yml",
 			UseTLSClient:   true,
 			ExpectedError:  ErrorMap["Unauthorized"],
 		},
 		{
 			Name:           `with correct basic auth and TLS`,
-			YAMLConfigPath: "testdata/tls_config_users.good.yml",
+			YAMLConfigPath: "testdata/web_config_users.good.yml",
 			UseTLSClient:   true,
 			Username:       "dave",
 			Password:       "dave123",
@@ -558,7 +649,7 @@ func TestUsers(t *testing.T) {
 		},
 		{
 			Name:           `with another correct basic auth and TLS`,
-			YAMLConfigPath: "testdata/tls_config_users.good.yml",
+			YAMLConfigPath: "testdata/web_config_users.good.yml",
 			UseTLSClient:   true,
 			Username:       "carol",
 			Password:       "carol123",
@@ -566,7 +657,7 @@ func TestUsers(t *testing.T) {
 		},
 		{
 			Name:           `with bad password and TLS`,
-			YAMLConfigPath: "testdata/tls_config_users.good.yml",
+			YAMLConfigPath: "testdata/web_config_users.good.yml",
 			UseTLSClient:   true,
 			Username:       "dave",
 			Password:       "bad",
@@ -574,7 +665,7 @@ func TestUsers(t *testing.T) {
 		},
 		{
 			Name:           `with bad username and TLS`,
-			YAMLConfigPath: "testdata/tls_config_users.good.yml",
+			YAMLConfigPath: "testdata/web_config_users.good.yml",
 			UseTLSClient:   true,
 			Username:       "nonexistent",
 			Password:       "nonexistent",
